@@ -6,38 +6,21 @@ http://www.biblioscape.com/rtf15_spec.htm
 This module is potentially compatible with RTF versions up to 1.9.1,
 but may not ignore all necessary control groups.
 """
-import string, re, itertools
+import string, re, itertools, struct
 
 from pyth import document
 from pyth.format import PythReader
+from pyth.encodings import symbol
 
 _CONTROLCHARS = set(string.ascii_letters + string.digits + "-*")
 _DIGITS = set(string.digits)
 
-# Maps Symbol typeface to Unicode, extracted from http://en.wikipedia.org/wiki/Symbol_(typeface)
-symbolTable = {
-    33: 33, 34: 8704, 35: 35, 36: 8707, 37: 37, 38: 38, 39: 8717, 40: 40, 41: 41, 42: 42, 43: 43, 44: 44, 45: 45, 46: 46, 
-    47: 47, 48: 48, 49: 49, 50: 50, 51: 51, 52: 52, 53: 53, 54: 54, 55: 55, 56: 56, 57: 57, 58: 58, 59: 59, 60: 60, 61: 61,
-    62: 62, 63: 63, 64: 8773, 65: 913, 66: 914, 67: 935, 68: 916, 69: 917, 70: 934, 71: 915, 72: 919, 73: 921, 74: 977, 
-    75: 922, 76: 923, 77: 924, 78: 925, 79: 927, 80: 928, 81: 920, 82: 929, 83: 931, 84: 932, 85: 933, 86: 962, 87: 937, 
-    88: 926, 89: 936, 90: 918, 91: 91, 92: 8756, 93: 93, 94: 8869, 95: 95, 96: 63717, 97: 945, 98: 946, 99: 967, 100: 948, 
-    101: 949, 102: 966, 103: 947, 104: 951, 105: 953, 106: 981, 107: 954, 108: 955, 109: 956, 110: 957, 111: 959, 112: 960, 
-    113: 952, 114: 961, 115: 963, 116: 964, 117: 965, 118: 982, 119: 969, 120: 958, 121: 968, 122: 950, 123: 123, 124: 124, 
-    125: 125, 126: 126, 160: 8364, 161: 978, 162: 697, 163: 8804, 164: 8260, 165: 8734, 166: 402, 167: 9827, 168: 9830, 
-    169: 9829, 170: 9824, 171: 8596, 172: 8592, 173: 8593, 174: 8594, 175: 8595, 176: 176, 177: 177, 178: 698, 179: 8805, 
-    180: 215, 181: 8733, 182: 8706, 183: 8226, 184: 247, 185: 8800, 186: 8801, 187: 8776, 188: 8230, 189: 9168, 190: 9135, 
-    191: 8629, 192: 8501, 193: 8465, 194: 8476, 195: 8472, 196: 8855, 197: 8853, 198: 8709, 199: 8745, 200: 8746, 201: 8835, 
-    202: 8839, 203: 8836, 204: 8834, 205: 8838, 206: 8712, 207: 8713, 208: 8736, 209: 8711, 210: 174, 211: 169, 212: 8482, 
-    213: 8719, 214: 8730, 215: 8901, 216: 172, 217: 8743, 218: 8744, 219: 8660, 220: 8656, 221: 8657, 222: 8658, 223: 8659,
-    224: 9674, 225: 12296, 226: 174, 227: 169, 228: 8482, 229: 8721, 230: 9115, 231: 9116, 232: 9117, 233: 9121, 234: 9122, 
-    235: 9123, 236: 9127, 237: 9128, 238: 9129, 239: 9130, 241: 12297, 242: 8747, 243: 8992, 244: 9134, 245: 8993, 246: 9118, 
-    247: 9119, 248: 9120, 249: 9124, 250: 9125, 251: 9126, 252: 9131, 253: 9132, 254: 9133}
 
 _CODEPAGES = {
     0: "cp1252",   # ANSI
     1: "cp1252",   # Default (this is wrong, but there is no right)
 
-    2: symbolTable,   # Symbol
+    2: "symbol",   # Symbol
     77: "mac-roman", # Mac Roman
 
     # Does Python have built-in support for these? What is it?
@@ -64,7 +47,7 @@ _CODEPAGES = {
     162: "cp1254", # Turkish
     163: "cp1258", # Vietnamese
     177: "cp1255", # Hebrew
-    178: "cp1256", # Arabic 
+    178: "cp1256", # Arabic
     186: "cp1257", # Baltic
     204: "cp1251", # Russian
     222: "cp874",  # Thai
@@ -73,8 +56,18 @@ _CODEPAGES = {
     255: "cp850",  # OEM
 }
 
+# All the ones named by number in my 2.6 encodings dir
+_CODEPAGES_BY_NUMBER = dict(
+    (x, "cp%s" % x) for x in (37, 1006, 1026, 1140, 1250, 1251, 1252, 1253, 1254, 1255,
+                              1256, 1257, 1258, 424, 437, 500, 737, 775, 850, 852, 855,
+                              856, 857, 860, 861, 862, 863, 864, 865, 866, 869, 874,
+                              875, 932, 949, 950))
 
-
+# Miscellaneous, incomplete
+_CODEPAGES_BY_NUMBER.update({
+   10000: "mac-roman",
+   10007: "mac-greek",
+})
 
 
 class BackslashEscape(Exception):
@@ -84,17 +77,18 @@ class BackslashEscape(Exception):
 class Rtf15Reader(PythReader):
 
     @classmethod
-    def read(self, source):
+    def read(self, source, errors='strict'):
         """
         source: A list of P objects.
         """
 
-        reader = Rtf15Reader(source)
+        reader = Rtf15Reader(source, errors)
         return reader.go()
 
 
-    def __init__(self, source):
+    def __init__(self, source, errors='strict'):
         self.source = source
+        self.errors = errors
         self.document = document.Document
 
 
@@ -107,8 +101,9 @@ class Rtf15Reader(PythReader):
 
         self.source.seek(0)
 
-        self.group = Group()
         self.charsetTable = None
+        self.charset = 'cp1252'
+        self.group = Group(self)
         self.stack = [self.group]
         self.parse()
         return self.build()
@@ -124,23 +119,29 @@ class Rtf15Reader(PythReader):
             if next in '\r\n':
                 continue
             if next == '{':
-                subGroup = Group(self.group, self.charsetTable)
+                subGroup = Group(self, self.group, self.charsetTable)
                 self.stack.append(subGroup)
+                subGroup.skip = self.group.skip
                 self.group = subGroup
             elif next == '}':
                 subGroup = self.stack.pop()
                 self.group = self.stack[-1]
-
                 subGroup.finalize()
+
                 if subGroup.specialMeaning == 'FONT_TABLE':
                     self.charsetTable = subGroup.charsetTable
                 self.group.content.append(subGroup)
 
+            elif self.group.skip:
+                # Avoid crashing on stuff we can't handle
+                # inside groups we don't care about anyway
+                continue
+
             elif next == '\\':
                 control, digits = self.getControl()
                 self.group.handle(control, digits)
             else:
-                self.group.char(unicode(next))
+                self.group.char(next)
 
 
     def getControl(self):
@@ -154,7 +155,7 @@ class Rtf15Reader(PythReader):
             if not next:
                 break
 
-            if first and next == '\\':
+            if first and next in '\\{}':
                 chars.extend("control_symbol")
                 digits.append(next)
                 break
@@ -191,7 +192,7 @@ class Rtf15Reader(PythReader):
 
     def build(self):
         doc = document.Document()
-       
+
         ctx = DocBuilder(doc)
 
         for bit in self.group.flatten():
@@ -220,7 +221,7 @@ class DocBuilder(object):
             self.block = document.Paragraph()
 
         self.block.content.append(
-            document.Text(self.propStack[-1].copy(), 
+            document.Text(self.propStack[-1].copy(),
                           [u"".join(self.run)]))
 
         self.run[:] = []
@@ -228,7 +229,7 @@ class DocBuilder(object):
 
     def cleanParagraph(self):
         """
-        Compress text runs, remove whitespace at start and end, 
+        Compress text runs, remove whitespace at start and end,
         skip empty blocks, etc
         """
 
@@ -243,9 +244,9 @@ class DocBuilder(object):
 
         for run in runs:
 
-            if run.content[0]: 
+            if run.content[0]:
                 hasContent = True
-            else: 
+            else:
                 continue
 
             # For whitespace-only groups, remove any property stuff,
@@ -291,7 +292,7 @@ class DocBuilder(object):
 
 
     def handle_Para(self, para):
-        
+
         self.flushParagraph()
 
         prevListLevel = self.listLevel
@@ -312,7 +313,7 @@ class DocBuilder(object):
         self.flushRun()
         self.propStack[-1].clear()
 
-    
+
     def handle_ReadableMarker(self, marker):
         self.flushRun()
         if marker.val:
@@ -331,7 +332,8 @@ class DocBuilder(object):
 
 class Group(object):
 
-    def __init__(self, parent=None, charsetTable=None):
+    def __init__(self, reader, parent=None, charsetTable=None):
+        self.reader = reader
         self.parent = parent
 
         if parent:
@@ -339,7 +341,7 @@ class Group(object):
             self.charset = self.parent.charset
         else:
             self.props = {}
-            self.charset = 'cp1252' # ?
+            self.charset = self.reader.charset
 
         self.specialMeaning = None
         self.skip = False
@@ -369,11 +371,11 @@ class Group(object):
 
 
     def char(self, char):
-        self.content.append(char)
+        self.content.append(char.decode(self.charset, self.reader.errors))
 
 
     def _finalize(self):
-        
+
         if self.destination:
             self.skip = True
 
@@ -381,7 +383,7 @@ class Group(object):
             self.skip = True
 
         if self.skip:
-            return       
+            return
 
         stuff = []
         i = 0
@@ -416,18 +418,42 @@ class Group(object):
         return stuff
 
 
+    # Header stuff
+    def handle_ansi(self): self.charset = self.reader.charset = 'cp1252'
+    def handle_mac(self): self.charset = self.reader.charset = 'mac-roman'
+    def handle_pc(self): self.charset = self.reader.charset = 'cp437'
+    def handle_pca(self): self.charset = self.reader.charset = 'cp850'
+
+    def handle_ansicpg(self, codepage):
+        codepage = int(codepage)
+        if codepage in _CODEPAGES_BY_NUMBER:
+            self.charset = self.reader.charset = _CODEPAGES_BY_NUMBER[codepage]
+        else:
+            raise ValueError("Unknown codepage %s" % codepage)
+
+
     def handle_fonttbl(self):
         self.specialMeaning = 'FONT_TABLE'
         self.charsetTable = {}
 
 
+    def _setFontCharset(self, charset=None):
+        if charset is None:
+            charset = self.reader.charset
+        # XXX Todo: Figure out a more graceful way to handle the fact that
+        # RTF font declarations can be in their own groups or not
+        if self.parent.charsetTable is not None:
+            self.parent.charsetTable[self.fontNum] = charset
+        else:
+            self.charsetTable[self.fontNum] = charset
+
     def handle_f(self, fontNum):
         if 'FONT_TABLE' in (self.parent.specialMeaning, self.specialMeaning):
             self.fontNum = int(fontNum)
+            self._setFontCharset()
         elif self.charsetTable is not None:
             self.charset = self.charsetTable[int(fontNum)]
 
-            
     def handle_fcharset(self, charsetNum):
         if 'FONT_TABLE' in (self.parent.specialMeaning, self.specialMeaning):
             # Theoretically, \fN should always be before \fcharsetN
@@ -435,12 +461,9 @@ class Group(object):
             # if it's not, and see if it happens in the real world.
             charset = _CODEPAGES.get(int(charsetNum))
 
-            # XXX Todo: Figure out a more graceful way to handle the fact that
-            # RTF font declarations can be in their own groups or not
-            if self.parent.charsetTable is not None:
-                self.parent.charsetTable[self.fontNum] = charset
-            else: 
-                self.charsetTable[self.fontNum] = charset
+            if charset is None:
+                raise ValueError("Unsupported charset %s" % charsetNum)
+            self._setFontCharset(charset)
 
 
     def handle_ansi_escape(self, code):
@@ -449,16 +472,12 @@ class Group(object):
         if isinstance(self.charset, dict):
             uni_code = self.charset.get(code)
             if uni_code is None:
-                char = '?'
+                char = u'?'
             else:
                 char = unichr(uni_code)
-            
 
         else:
-            try:
-                char = chr(code).decode(self.charset)
-            except UnicodeDecodeError:
-                char = '?'
+            char = chr(code).decode(self.charset, self.reader.errors)
 
         self.content.append(char)
 
@@ -470,7 +489,16 @@ class Group(object):
 
 
     def handle_u(self, codepoint):
-        self.content.append(unichr(int(codepoint)))
+        codepoint = int(codepoint)
+        try: 
+            char = unichr(codepoint)
+        except ValueError:
+            if self.reader.errors == 'replace':
+                char = '?'
+            else:
+                raise
+
+        self.content.append(char)
         self.content.append(Skip(self.props.get('unicode_skip', 1)))
 
 
